import copy, math

import pyglet
from pyglet.gl import *

class Level:
    """Represents a 2D game level"""
    
    WALL = 0
    FLOOR = (255 << 16) + (255 << 8) + 255
    CEILING = (255 << 16) + (255 << 8) + 256
    SPAWN = 255
    GOAL = 255 << 8
    TRODDEN = (255 << 8) + 255
    KEY = (255 << 16) + 255
    DOOR = 255 << 16
    STAR = (255 << 16) + (255 << 8)
    
    texture = {
        FLOOR: pyglet.resource.texture("tex/floor.png"),
        WALL: pyglet.resource.texture("tex/wall.png"),
        CEILING: pyglet.resource.texture("tex/ceiling.png"),
        DOOR: pyglet.resource.texture("tex/door.png"),
        STAR: pyglet.resource.texture("tex/star.png"),
        TRODDEN: pyglet.resource.texture("tex/trodden.png"),
    }

    def __init__(self):
        self.height = 0
        self.width = 0
        self.spawnX = None
        self.spawnY = None
        self.spawnHeading = None
        self.stars_max = 0
        self.original = []
        self.batch = pyglet.graphics.Batch()
        self.entity_list = []
        
        self.clear()
    
    def cache(self):
        self.batch = pyglet.graphics.Batch()
        for x in xrange(0, self.width):
            for y in xrange(0, self.height):
                if self.data[x][y] == Level.FLOOR:
                    # Add FLOOR
                    self.batch.add(4, GL_QUADS,
                        pyglet.graphics.TextureGroup(Level.texture[Level.FLOOR]),
                        ('v3f', (
                                x,     y, 0,
                            x + 1,     y, 0,
                            x + 1, y + 1, 0,
                                x, y + 1, 0
                        )), ('t2f', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                    )
                    # Add CEILING
                    self.batch.add(4, GL_QUADS,
                        pyglet.graphics.TextureGroup(Level.texture[Level.CEILING]),
                        ('v3f', (
                            x + 1,     y, 2,
                                x,     y, 2,
                                x, y + 1, 2,
                            x + 1, y + 1, 2,
                        )), ('t2f', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                    )
                elif self.data[x][y] == Level.TRODDEN:
                    # Add TRODDEN
                    self.batch.add(4, GL_QUADS,
                        pyglet.graphics.TextureGroup(Level.texture[Level.TRODDEN]),
                        ('v3f', (
                                x,     y, 0,
                            x + 1,     y, 0,
                            x + 1, y + 1, 0,
                                x, y + 1, 0
                        )), ('t2f', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                    )
                    # Add CEILING
                    self.batch.add(4, GL_QUADS,
                        pyglet.graphics.TextureGroup(Level.texture[Level.CEILING]),
                        ('v3f', (
                            x + 1,     y, 2,
                                x,     y, 2,
                                x, y + 1, 2,
                            x + 1, y + 1, 2,
                        )), ('t2f', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                    )
                elif self.data[x][y] == Level.WALL:
                    # Add WALLs
                    if x > 0 and self.data[x-1][y] != Level.WALL:
                        self.batch.add(4, GL_QUADS,
                            pyglet.graphics.TextureGroup(Level.texture[Level.WALL]),
                            ('v3f', (
                                x, y + 1, 0,
                                x,     y, 0,
                                x,     y, 2,
                                x, y + 1, 2,
                            )), ('t2f', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                        )
                    if x < self.width-1 and self.data[x+1][y] != Level.WALL:
                        self.batch.add(4, GL_QUADS,
                            pyglet.graphics.TextureGroup(Level.texture[Level.WALL]),
                            ('v3f', (
                                x + 1,     y, 0,
                                x + 1, y + 1, 0,
                                x + 1, y + 1, 2,
                                x + 1,     y, 2,
                            )), ('t2f', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                        )
                    if y > 0 and self.data[x][y-1] != Level.WALL:
                        self.batch.add(4, GL_QUADS,
                            pyglet.graphics.TextureGroup(Level.texture[Level.WALL]),
                            ('v3f', (
                                    x,     y, 0,
                                x + 1,     y, 0,
                                x + 1,     y, 2,
                                    x,     y, 2,
                            )), ('t2f', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                        )
                    if y < self.height-1 and self.data[x][y+1] != Level.WALL:
                        self.batch.add(4, GL_QUADS,
                            pyglet.graphics.TextureGroup(Level.texture[Level.WALL]),
                            ('v3f', (
                                x + 1, y + 1, 0,
                                    x, y + 1, 0,
                                    x, y + 1, 2,
                                x + 1, y + 1, 2,
                            )), ('t2f', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                        )
                elif self.data[x][y] == Level.DOOR:
                    self.batch.add(16, GL_QUADS,
                        pyglet.graphics.TextureGroup(Level.texture[Level.DOOR]),
                        ('v3f', (
                            x, y + 1, 0,
                            x,     y, 0,
                            x,     y, 2,
                            x, y + 1, 2,
                                
                            x + 1,     y, 0,
                            x + 1, y + 1, 0,
                            x + 1, y + 1, 2,
                            x + 1,     y, 2,
                            
                                x,     y, 0,
                            x + 1,     y, 0,
                            x + 1,     y, 2,
                                x,     y, 2,
                            
                            x + 1, y + 1, 0,
                                x, y + 1, 0,
                                x, y + 1, 2,
                            x + 1, y + 1, 2,
                        )), ('t2i', (
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                            0, 0,
                            1, 0,
                            1, 1,
                            0, 1,
                        ))
                    )
        
    def clear(self):
        """Reset level to empty"""

        self.data = []
        self.original = []
        for x in xrange(0, self.width):
            self.data.append([])
            self.original.append([])
            for y in xrange(0, self.height):
                self.data[x].append(0)
                self.original[x].append(0)
                
    def draw(self):
        """Draw the level"""
        
        self.batch.draw()
    
    def load_from_file(self, filename):
        """Load the level from an image file"""
        
        image = pyglet.resource.image(filename)
        
        self.height = image.height
        self.width = image.width
        
        self.clear()
        
        raw_image = image.get_image_data()
        format = 'RGB'
        pitch = raw_image.width * len(format)
        data = raw_image.get_data(format, pitch)
    
        for x in xrange(0, self.width):
            for y in xrange(0, self.height):
                r = ord(data[(self.width * y + x) * 3])
                g = ord(data[(self.width * y + x) * 3 + 1])
                b = ord(data[(self.width * y + x) * 3 + 2])
                self.original[x][y] = (r << 16) + (g << 8) + b
                self.special_tile(x, y, self.original[x][y], self.original)
                self.data[x][y] = self.original[x][y]
        
        self.original = copy.deepcopy(self.data)
        
        self.cache()

    def reset(self):
        for x in xrange(0, self.width):
            for y in xrange(0, self.height):
                self.data[x][y] = self.original[x][y]
    
    def special_tile(self, x, y, tile, data):
        if data[x][y] == Level.SPAWN:
            self.spawnX = x + 0.5
            self.spawnY = y + 0.5
            self.spawnHeading = 0
            data[x][y] = Level.FLOOR
        elif data[x][y] == Level.GOAL:
            self.entity_list.append(("Goal", x + .5, y + .5))
            data[x][y] = Level.FLOOR
        elif data[x][y] == Level.KEY:
            self.entity_list.append(("Key", x + .5, y + .5))
            data[x][y] = Level.FLOOR
        elif data[x][y] == Level.STAR:
            self.entity_list.append(("Star", x + .5, y + .5))
            self.stars_max += 1
            data[x][y] = Level.FLOOR
